#include <FreeRTOS.h>
#include <task.h>
#include <portable.h>
#include <queue.h>
#include <string.h>

#include <libopencm3/stm32/gpio.h>
#include <libopencm3/stm32/rcc.h>
#include <libopencm3/stm32/exti.h>
#include <libopencm3/stm32/usart.h>
#include <libopencm3/cm3/nvic.h>

#include "lib/mini-printf.h"
#include "lib/gpio_utils.h"
#include "lib/serial_utils.h"
#include "lib/ad98xx.h"

#ifdef __GNUC__
#define UNUSED(x) UNUSED_##x __attribute__((__unused__))
#else
#define UNUSED(x) UNUSED_##x
#endif

#define DDS_REF ((float)125e6)

//add stack-overflow hook
extern void vApplicationStackOverflowHook(xTaskHandle *pxTask, signed portCHAR *pcTaskName);

// our tasks
void task_onboardled(void *args);
void task_uart_tx(void *args);
void task_uart_rx(void *args);
static QueueHandle_t uart_tx_q;

/*
 * Partiamo dall'esempio `serialpc` e aggiungiamo il controllo del DDS sulle porte PB_15 e PB_13. Queste porte
 * sono anche utilizzabili come SPI2 (rispettivamente MOSI e SCK), ma in questo esempio facciamo bit-banging.
 * Scegliamo però le porte "giuste" in modo che quando faremo la versione SPI abbiamo tutto fatto giusto.
 */

static void gpio_setup(void)
{

	// built-in LED
	rcc_periph_clock_enable(RCC_GPIOC);
	gpio_set_mode(GPIOC, GPIO_MODE_OUTPUT_2_MHZ, GPIO_CNF_OUTPUT_PUSHPULL, GPIO13);

	rcc_periph_clock_enable(RCC_GPIOA);
	rcc_periph_clock_enable(RCC_GPIOB);

    // Serial comunication
    rcc_periph_clock_enable(RCC_USART1);
    gpio_set_mode(GPIOA, GPIO_MODE_OUTPUT_50_MHZ,
            GPIO_CNF_OUTPUT_ALTFN_PUSHPULL,
            GPIO_USART1_TX);
    usart_set_baudrate(USART1, 38400);
    usart_set_databits(USART1,8);
    usart_set_stopbits(USART1,USART_STOPBITS_1);
    usart_set_mode(USART1,USART_MODE_TX_RX);
    usart_set_parity(USART1,USART_PARITY_NONE);
    usart_set_flow_control(USART1,USART_FLOWCONTROL_NONE);
    usart_enable(USART1);
}


static void uart_puts(const char *s) {
    for(; *s; ++s) {
        usart_send_blocking(USART1, *s);
    }
}

ad98_dds dds = {2, GPIOB, GPIO13, GPIO14, GPIO15, GPIO12};

int main(void)
{
    rcc_clock_setup_in_hse_8mhz_out_72mhz();
	gpio_setup();
    ad98_init(dds);
    uart_puts("READY\n");

    uart_tx_q = xQueueCreate(20, sizeof(char*));
	xTaskCreate(task_onboardled, "ONBOARD", 50, NULL, configMAX_PRIORITIES - 1,
	            NULL);
	xTaskCreate(task_uart_tx, "TX", 50, NULL, configMAX_PRIORITIES - 1,
	            NULL);
	// xTaskCreate(task_uart_rx, "RX", 50, NULL, configMAX_PRIORITIES - 1,
	//             NULL);
	vTaskStartScheduler();
	for (;;) {
	}

	return 0;
}

int boardLedDuty = 1000;
void task_onboardled(void *UNUSED(args))
{
    float dds_frequency = 200.0;
	while (1) {
        q_printf(uart_tx_q, 16, "f=%d\n", (int)dds_frequency);
        gpio_clear(GPIOC, GPIO13); /* LED on */
        vTaskDelay(pdMS_TO_TICKS(boardLedDuty));

		gpio_set(GPIOC, GPIO13); /* LED off */
		vTaskDelay(pdMS_TO_TICKS(500));

        if(dds_frequency > 12000) {
            dds_frequency = 200.0;
        }
        ad98_set_frequency(dds, (double)dds_frequency);
        dds_frequency *= 1.3;
	}
}

void task_uart_tx(void *UNUSED(args)) {
    while (1) {
        char *buf, *buf2;
        if(xQueueReceive(uart_tx_q, (void*)&buf, pdMS_TO_TICKS(5000)) != pdPASS)
            continue;
        buf2 = buf; // mi devo tenere il puntatore originale senno' dopo non posso fare free
        for(; *buf != '\0'; buf++) {
            while(!usart_get_flag(USART1, USART_SR_TXE))
                taskYIELD();
            usart_send_blocking(USART1, *buf);
        }
        vPortFree(buf2);
    }
}

void task_uart_rx(void *UNUSED(args)) {
    char cmd[32];
    unsigned char i = 0;
    while (1) {
        if (i >= 30) { // avoid overflows
            i = 0;
        }
        cmd[i] = usart_recv_blocking(USART1);
        if (cmd[i] == '\r' || cmd[i] == '\n') {
            gpio_toggle(GPIOB, GPIO11);
            cmd[i] = '\0'; // replace newline with end-of-string
            if (strcmp(cmd, "GET") == 0) { //check valid commands
                char *buf = (void *)pvPortMalloc(16);
                if (buf != NULL) {
                    snprintf(buf, 16, "-> %d\n", boardLedDuty);
                    if (xQueueSend(uart_tx_q, (void *)&buf, 0) != pdTRUE) {
                        vPortFree(buf);
                    }
                }
            } // GET
            i=0;
        } // newline
        else {
            i++;
        }
    }
}

// three "small flashes" (an S in Morse code) on onboard led
//
// questa funzione viene chiamata in caso di stackoverflow su uno dei task; e'
// quindi utile metterci un "segnale" riconoscibile per debug. In questo caso 3
// flash rapidi.
//
// siccome "tarare" la dimensione dello stack necessario e' difficile, con
// questa funzione si puo' andare "a tentoni" e ridurre la dimensione finche'
// non si ottengono crash
//
// in questa funzione, a quanto pare (ma non ho trovato doc) non si possono
// usare cose come vTaskDelay. Siamo quindi privi di RTOS e per fare gli sleep
// dobbiamo fare i for "all'antica"

void vApplicationStackOverflowHook(xTaskHandle *UNUSED(pxTask),
        signed portCHAR *UNUSED(pcTaskName))
{
    int i;
    for (;;) {
        for (char c = 0; c < 3; c++) {
            gpio_clear(GPIOC, GPIO13); /* LED on */
            for (i = 0; i < 1000000; i++)   /* small delay */
                __asm__("nop");

            gpio_set(GPIOC, GPIO13);   /* LED off */
            for (i = 0; i < 3000000; i++) { /* a bit more delay */
                __asm__("nop");
            }
        }
        for (i = 0; i < 10000000; i++) {    /* long delay */
            __asm__("nop");
        }
    }
}

// vim: set fdm=marker:
