#include <FreeRTOS.h>
#include <task.h>
#include <queue.h>
#include <libopencm3/stm32/gpio.h>
#include <libopencm3/stm32/rcc.h>
#include <libopencm3/stm32/exti.h>
#include <libopencm3/cm3/nvic.h>

#ifdef __GNUC__
#define UNUSED(x) UNUSED_##x __attribute__((__unused__))
#else
#define UNUSED(x) UNUSED_##x
#endif

//add stack-overflow hook
extern void vApplicationStackOverflowHook(xTaskHandle *pxTask, signed portCHAR *pcTaskName);

// our tasks
void task_onboardled(void *args);
void task_buttonled(void *args);
static QueueHandle_t btn_q; // coda dall'ISR del pulsante a task_buttonled

/*
 * applicazione di esempio. controlla un rotary encoder e 3 led
 * un led (quello built-in) fa on/off con duty cycle variabile
 *
 * altri due led (porte A8 e B11) si accendono seguendo il rotary encoder (porte A1 e A0) è premuto (mandato a
 * VCC): quando si ruota in un verso, il led rosso è fermo e quello verde cambia; nella direzione opposta
 * avviene l'inverso.
 * Il rotary controlla inoltre il duty cycle del led on-board. È possibile quindi testare il tutto anche senza
 * montare i LED, ma sarà tendenzialmente più difficile debuggare.
 *
 * OSSERVAZIONI
 * - Il tutto è leggermente instabile perché il codice che controlla il rotary è molto semplice e non verifica
 *   gli errori
 * - Il debouncing è fatto in hw (vedi sotto), che può pure andar bene, ma vedere come si fa il debouncing in
 *   sw sembra una cosa utile
 *
 *
 * MONTAGGIO
 * - un led va messo tra la porta A8, una resistenza (330ohm) e la massa
 * - un led va messo tra la porta B11, una resistenza (330ohm) e la massa
 * - il rotary encoder vuole il VCC sul pin centrale, porte A1 e A0 ai lati.
 *   Per fare "debouncing" si usano dei condensatori da 0.1uF tra ciascun pin "laterale" e il VCC. Ovvero ci
 *   vuole un condensatore tra A1 e VCC e uno tra A0 e VCC. Il tutto è dunque in parallelo rispetto al rotary.
 */

static void gpio_setup(void)
{

	// built-in LED
	rcc_periph_clock_enable(RCC_GPIOC);
	gpio_set_mode(GPIOC, GPIO_MODE_OUTPUT_2_MHZ, GPIO_CNF_OUTPUT_PUSHPULL,
	              GPIO13);

	// ext. rotary button: due pin per la direzione (A3 e A4). Metto l'interrupt solo su uno dei due (B1)
	rcc_periph_clock_enable(RCC_GPIOA);
	gpio_set_mode(GPIOA, GPIO_MODE_INPUT, GPIO_CNF_INPUT_PULL_UPDOWN, GPIO0|GPIO1);

    exti_select_source(EXTI0, GPIOA);
    exti_set_trigger(EXTI0, EXTI_TRIGGER_RISING);
    exti_enable_request(EXTI0);
    nvic_enable_irq(NVIC_EXTI0_IRQ); // occhio! guarda tabella 11-2 a pag. 209 o ti perdi su quale NVIC_EXTI usare

	// ext. led (red)
	gpio_set_mode(GPIOA, GPIO_MODE_OUTPUT_2_MHZ, GPIO_CNF_OUTPUT_PUSHPULL, GPIO8);
	// ext. led (green)
	rcc_periph_clock_enable(RCC_GPIOB);
	gpio_set_mode(GPIOB, GPIO_MODE_OUTPUT_2_MHZ, GPIO_CNF_OUTPUT_PUSHPULL, GPIO11);
}


int main(void)
{
    rcc_clock_setup_in_hse_8mhz_out_72mhz();
	gpio_setup();
    gpio_set(GPIOA, GPIO8);
    gpio_clear(GPIOB, GPIO11);

    btn_q = xQueueCreate(10, sizeof(char));
	xTaskCreate(task_buttonled, "BUTTONLED", 50, NULL, configMAX_PRIORITIES - 1,
	            NULL);
	xTaskCreate(task_onboardled, "ONBOARD", 50, NULL, configMAX_PRIORITIES - 1,
	            NULL);
	vTaskStartScheduler();
	for (;;) {
	}

	return 0;
}

int boardLedDuty = 1000;
void task_onboardled(void *UNUSED(args))
{
	while (1) {
		gpio_clear(GPIOC, GPIO13); /* LED on */
		vTaskDelay(pdMS_TO_TICKS(boardLedDuty));

		gpio_set(GPIOC, GPIO13); /* LED off */
		vTaskDelay(pdMS_TO_TICKS(500));
	}
}

void task_buttonled(void *UNUSED(args))
{
    while (1) {
        char increasing;
        if(xQueueReceive(btn_q, &increasing, pdMS_TO_TICKS(5000)) == pdPASS)
        {
            if(increasing) {
                gpio_toggle(GPIOA, GPIO8);
            } else {
                gpio_toggle(GPIOB, GPIO11);
            }
        }
    }
}


// gestisce l'interrupt del bottone. Viene chiamato quando il pulsante viene premuto, non quando viene
// rilasciato
void exti0_isr(void)
{
    char v = gpio_get(GPIOA, GPIO1);
    exti_reset_request(EXTI0);
    if(v) { // decrease
        if(boardLedDuty < 220)
            boardLedDuty = 30;
        else
            boardLedDuty -= 200;
        xQueueSend(btn_q, (char) 0, 0);
    } else {
        if(boardLedDuty >= 2300)
            boardLedDuty = 2500;
        else
            boardLedDuty += 200;
        xQueueSend(btn_q, (char) 1, 0);
    }
}

// three "small flashes" (an S in Morse code) on onboard led
//
// questa funzione viene chiamata in caso di stackoverflow su uno dei task; e'
// quindi utile metterci un "segnale" riconoscibile per debug. In questo caso 3
// flash rapidi.
//
// siccome "tarare" la dimensione dello stack necessario e' difficile, con
// questa funzione si puo' andare "a tentoni" e ridurre la dimensione finche'
// non si ottengono crash
//
// in questa funzione, a quanto pare (ma non ho trovato doc) non si possono
// usare cose come vTaskDelay. Siamo quindi privi di RTOS e per fare gli sleep
// dobbiamo fare i for "all'antica"

void vApplicationStackOverflowHook(xTaskHandle *UNUSED(pxTask),
        signed portCHAR *UNUSED(pcTaskName))
{
    int i;
    for (;;) {
        for (char c = 0; c < 3; c++) {
            gpio_clear(GPIOC, GPIO13); /* LED on */
            for (i = 0; i < 1000000;
                    i++)                       /* small delay */
                __asm__("nop");

            gpio_set(GPIOC, GPIO13);       /* LED off */
            for (i = 0; i < 3000000; i++) { /* a bit more delay */
                __asm__("nop");
            }
        }
        for (i = 0; i < 10000000; i++) { /* long delay */
            __asm__("nop");
        }
    }
}

// vim: set fdm=marker:
